/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Mozilla Inspector Module.
 *
 * The Initial Developer of the Original Code is
 * The Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Rob Campbell <rcampbell@mozilla.com> (original author)
 *   Paul Rouget <paul@mozilla.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

body {
  margin: 0;
  overflow: hidden;
}

#close-button {
  background-image: url("KUI-close.png");
  border: none;
  padding: 0;
  width: 24px;
  height: 24px;
  position: fixed;
  top: 12px;
  right: 12px;
  z-index: 1;
  cursor: pointer;
}

.veil {
  background-color: rgba(0, 0, 0, 0.5);
}

.veil, #veil-middlebox, #veil-transparentbox {
  -moz-transition: 0.1s;
  -moz-transition-timing-function: linear;
}

#veil-container {
  position: absolute;
  top: 0;
  left: 0;
  width: 100%;
  height: 100%;
}

#veil {
  width: 100%;
  height: 100%;
  display: -moz-box;
  -moz-box-orient: vertical;
}

#veil-topbox, #veil-bottombox {
  width: 100%;
}

#veil-bottombox {
  -moz-box-flex: 1;
}

#veil-middlebox {
  display: -moz-box;
  -moz-box-orient: horizontal;
}

#veil-leftbox, #veil-rightbox {
  height: 100%;
}

#veil-rightbox {
  -moz-box-flex: 1;
}

#veil {
  vertical-align: top;
}

